/*
  WiFi RC Transmitter
  2016-09-22 T. Nakagawa
*/

#include <EEPROM.h>
#include <ESP8266WiFi.h>
#include <Ticker.h>
#include <WiFiClient.h>
#include <WiFiUDP.h>
#include <limits.h>

#define LED_PIN 4
#define PPM_PIN 5
#define AP_SSID "AP0010"
#define AP_PASS "drondron"
#define HOST "192.168.4.1"
#define PORT 12345
#define CHANNELS 6
#define CALIB_CH0 0
#define CALIB_CH1 1
#define CALIB_THRESHOLD (1750 * 80)

IPAddress Host;
Ticker Flasher;
Ticker Timeout;
Ticker Watcher;
WiFiUDP Udp;
volatile boolean PpmInitialized = false;
int PpmCount = -1;
unsigned int PpmTime;
unsigned int PpmWidth[CHANNELS];
unsigned int PpmMin[CHANNELS];
unsigned int PpmNeu[CHANNELS];
unsigned int PpmMax[CHANNELS];

void FlushLed() {
  static boolean flag = false;
  digitalWrite(LED_PIN, flag ? LOW : HIGH);
  flag = !flag;
}

void PpmInitialize() {
  static int state = 0;
  static unsigned long start;

  switch (state) {
  case 0:  // Enter the calibration mode.
    if (PpmWidth[CALIB_CH0] >= CALIB_THRESHOLD && PpmWidth[CALIB_CH1] >= CALIB_THRESHOLD) {
      for (int i = 0; i < CHANNELS; i++) {
        PpmMin[i] = UINT_MAX;
        PpmMax[i] = 0;
      }
      state = 1;
      Flasher.attach_ms(500, FlushLed);
      start = millis();
      return;
    }
    break;
  case 1:  // Adjust min and max (20 sec.).
    if (millis() - start >= 20000) {
      state = 2;
      Flasher.attach_ms(250, FlushLed);
      start = millis();
      return;
    }
    for (int i = 0; i < CHANNELS; i++) {
      if (PpmWidth[i] < PpmMin[i]) {
        PpmMin[i] = PpmWidth[i];
      }
      if (PpmWidth[i] > PpmMax[i]) {
        PpmMax[i] = PpmWidth[i];
      }
    }
    return;
  case 2:  // Adjust neutral and store the parameters (10 sec.).
    if (millis() - start >= 10000) {
      for (int i = 0; i < CHANNELS; i++) {
        PpmNeu[i] = PpmWidth[i];
      }
      EEPROM.put<unsigned int [CHANNELS]>(sizeof(unsigned int [CHANNELS]) * 0, PpmMin);
      EEPROM.put<unsigned int [CHANNELS]>(sizeof(unsigned int [CHANNELS]) * 1, PpmNeu);
      EEPROM.put<unsigned int [CHANNELS]>(sizeof(unsigned int [CHANNELS]) * 2, PpmMax);
      EEPROM.commit();
      state = 3;
      Flasher.detach();
      digitalWrite(LED_PIN, HIGH);
      break;
    }
    return;
  }

  EEPROM.get<unsigned int [CHANNELS]>(sizeof(unsigned int [CHANNELS]) * 0, PpmMin);
  EEPROM.get<unsigned int [CHANNELS]>(sizeof(unsigned int [CHANNELS]) * 1, PpmNeu);
  EEPROM.get<unsigned int [CHANNELS]>(sizeof(unsigned int [CHANNELS]) * 2, PpmMax);
  // Validate the parameters.
  for (int i = 0; i < CHANNELS; i++) {
    if (!(750 * 80 <= PpmMin[i] && PpmMin[i] < PpmNeu[i] && PpmNeu[i] < PpmMax[i] && PpmMax[i] <= 2250 * 80)) {
      PpmMin[i] = 1000 * 80;
      PpmNeu[i] = 1500 * 80;
      PpmMax[i] = 2000 * 80;
    }          
  }
  PpmInitialized = true;
}

void SendData() {
  unsigned char buf[2 + 2 * CHANNELS];

  buf[0] = CHANNELS;
  buf[1] = 0;
  for (int i = 0; i < CHANNELS; i++) {
    unsigned short val;
    if (PpmWidth[i] < PpmNeu[i]) {
      if (PpmWidth[i] < PpmMin[i]) {
        PpmWidth[i] = PpmMin[i];
      }
      val = map(PpmWidth[i], PpmMin[i], PpmNeu[i], 0, 32768);
    } else{
      if (PpmWidth[i] > PpmMax[i]) {
        PpmWidth[i] = PpmMax[i];
      }
      val = map(PpmWidth[i], PpmNeu[i], PpmMax[i], 32768, 65535);
    }
    buf[2 + 2 * i + 0] = (val >> 8);
    buf[2 + 2 * i + 1] = (val & 0xff);
  }
  unsigned char chksum = 0;
  for (int i = 0; i < 2 + 2 * CHANNELS; i++) {
    chksum += buf[i];
  }
  buf[1] = -chksum;

  if (Udp.beginPacket(Host, PORT)) {
    Udp.write(buf, 2 + 2 * CHANNELS);
    Udp.endPacket();
  }
}

void TimeoutInterruption() {
  if (PpmCount == CHANNELS) {
    if (!PpmInitialized) {
      PpmInitialize();
    } else {
      SendData();
    }
  }
  PpmCount = -1;
}

void PpmInterruption() {
  const unsigned int currentTime = ESP.getCycleCount();
  if (PpmCount >= 0 && PpmCount < CHANNELS) {
    PpmWidth[PpmCount] = currentTime - PpmTime;
  }
  if (PpmCount <= CHANNELS) {
    PpmCount++;
  }
  PpmTime = currentTime;
  Timeout.once_ms(5, TimeoutInterruption);
}

void CheckWifi() {
  if (WiFi.status() != WL_CONNECTED) {
    InitWifi();
  }
}

void InitWifi() {
  Watcher.detach();
  Flasher.attach_ms(100, FlushLed);

  for (; ; ) {
    WiFi.mode(WIFI_STA);
    WiFi.begin(AP_SSID, AP_PASS);
    for (int i = 0; i < 300; i++) {
      if (WiFi.status() == WL_CONNECTED) {
        goto success;
      }
      delay(100);
    }
  }
success:

  Flasher.detach();
  digitalWrite(LED_PIN, HIGH);
  Watcher.attach(1, CheckWifi);
}

void setup() {
  pinMode(LED_PIN, OUTPUT);
  digitalWrite(LED_PIN, LOW);
  pinMode(PPM_PIN, INPUT);

  EEPROM.begin(sizeof(unsigned int [CHANNELS]) * 3);

  attachInterrupt(PPM_PIN, PpmInterruption, FALLING);
  while (!PpmInitialized) {
    delay(100);
  }

  Host.fromString(HOST);
  InitWifi();
}

void loop() {
}

